# -*- coding: UTF-8 -*-

import json
import os
import sys
import time
import traceback

sys.path[0] = os.path.normpath('%s/../../' % sys.path[0])

from bottle import request, response, HTTPResponse, Bottle
from common import logger, config, filesmodify, parameters, rest_state
from functools import wraps
from httplib import INTERNAL_SERVER_ERROR, UNAUTHORIZED
from upload.node.node import DBInserterConfig, QueueInserterConfig


class RestService(Bottle):

    def __init__(self, isCash=True):
        super(RestService, self).__init__()
        self.__readConfig()
        self.log = logger.getUploadLogger('rest')
        self.restLogName = 'dictLoadRestBackup'
        self.isCash = isCash
        self.restServiceHost = '0.0.0.0'
        self.route('/aifexchange/loaddicts', method='POST', callback=self.__restStatus)

    def start(self):
        self.log.info('Запуск REST-сервиса загрузки справочников, host = %s, port = %s' % (self.restServiceHost, self.restServicePort))
        try:
            self.install(self.__loggingRequest)
            self.run(host=self.restServiceHost, port=self.restServicePort, quiet=True)
        except Exception as e:
            self.log.error(u'Ошибка: %s' % e)
            self.log.error(traceback.format_exc())

    def __readConfig(self):
        self.primitiveAuth = config.getBool('REST_UPLOAD', 'authorize', True)
        self.restLogin = config.getString('REST_UPLOAD', 'login', 'admin')
        self.restPasswd = config.getString('REST_UPLOAD', 'passwd', 'admin')
        self.restServicePort = config.getInt('REST_UPLOAD', 'port', 28080)
        self.shopCodeIsShopId = config.getBool('REST_UPLOAD', 'shopCodeIsShopId', True)
        self.maxCountBackupRequests = config.getInt('REST_UPLOAD', 'maxCountBackupRequests', 0)
        self.actualCountDaysForBackupRequests = config.getInt('REST_UPLOAD', 'actualCountDaysForBackupRequests', 3)
        self.defaultStatusServer = config.getString('CS', 'defaultStatusServer', 'localhost:8080')
        self.defaultNesServer = config.getString('CS', 'defaultNesServer', 'localhost:28081')

    def __authorization(func):  # функция
        """Проверка авторизации"""

        def check_auth(self):  # обёртка wrapper
            if self.primitiveAuth:
                if request.auth and request.auth == (self.restLogin, self.restPasswd):
                    return func(self)
                else:
                    raise HTTPResponse("Unauthorized", status=UNAUTHORIZED)

        return check_auth

    @__authorization
    def __restStatus(self):
        """Обработка входящего запроса aifexchange/loaddicts с провекой авторизации"""
        response.content_type = 'application/json'
        body = None
        dictionaryId = request.query.get('dictionaryid', None)
        node = None
        try:
            if request.content_length > 0 or request.chunked:
                body = json.loads(request.body.read(), encoding='utf8', object_hook=self.__filterKeys)
            if not isinstance(body, list):
                raise Exception('Body is not array.')
            if len(body) == 0:
                body = None
                raise Exception('Array is empty.')
            node, result = self.__load(body, shopCode=request.query.get('shopcode', 'All'),
                                       dictionaryId=dictionaryId, statusServer=request.query.get('addressstatusserver', None),
                                       directLoad="directload" in request.query)

        except Exception as e:
            errorStr = u''
            errorStr_ru = u''
            if isinstance(body, list):
                errorStr = u'Error loading dicts. %s' % str(e)
                errorStr_ru = u'Ошибка загрузки справочника. %s' % str(e)
            else:
                errorStr = u'Incorrect request. %s' % str(e)
                errorStr_ru = u'Неправильный запрос. %s' % str(e)
            self.log.error(errorStr)
            self.log.error(traceback.format_exc())
            response.status = INTERNAL_SERVER_ERROR
            result = {u'success': False, u'info': errorStr, u'info_ru': errorStr_ru}
        if node:
            rest_state.sendDictState(node, result['info'], result['info_ru'], rest_state.STATUS_READY if result['success'] else rest_state.STATUS_ERROR,
                                     dictionaryId)

        return json.dumps(result, ensure_ascii=False)

    @staticmethod
    def __filterKeys(d):
        """используется в object_hook - Optional - функции, скорее всего, чтобы избежать None"""
        lowerD = {}
        for k in d.keys():
            if d[k] is not None and d[k] != []:
                lowerD[k.lower()] = d[k]
        return lowerD

    def __load(self, dictObjects, shopCode='All', dictionaryId=None, statusServer=None, directLoad=False):
        """Загрузка справочника"""
        self.log.info("Загрузка %s" % ("напрямую в БД" if directLoad else "через очередь"))
        node = EmptyNode(self.isCash, shopCode, directLoad, statusServer, dictionaryId).getNode()
        res, shopCode = self.__checkShopCode(shopCode, node)
        if not res:
            return None, {u'success': False, 'info': 'Shop with code = %s does not found.' % shopCode, 'info_ru': u'Магазин с кодом = %s не найден.' % shopCode}
        self.log.info('status server from request=%s' % statusServer)

        plugin = __import__('upload.inserter.%s' % node.inserter.name, fromlist='Inserter')
        cl = getattr(plugin, 'Inserter')
        inserter = cl(node)
        inserter.disableProgressFile = directLoad
        inserter.source = "exchangers-rest-upload"

        try:
            if node.currentConverter.hotKeyByBarcode:
                for item in dictObjects:
                    if 'addHotKey' == item['command'] and not 'bybarcode' in item['hotkey']:
                        item['hotkey']['bybarcode'] = node.currentConverter.hotKeyByBarcode
            result = inserter.start()
            if result:
                n = len(dictObjects)
                k = n / node.pack_size
                if n % node.pack_size > 0:
                    k += 1
                for i in range(k):
                    inserter.add_package(dictObjects[i * node.pack_size:(i * node.pack_size) + node.pack_size])
                result = inserter.stop()
                if result:
                    return node, {u'success': result, 'info': u'OK', u'info_ru': 'OK'}
                else:
                    return node, {u'success': result, 'info': u'Can not commit to queue.', 'info_ru': u'Не возможно положить в очередь объекты'}
            else:
                return node, {u'success': result, 'info': u'Can not start for load to queue.', 'info_ru': u'Не возможно начать загрузку в очередь.'}
        except Exception as e:
            self.log.error(u'Ошибка: {0}'.format(unicode(e)))
            self.log.error(traceback.format_exc())
            return node, {u'success': False, 'info': u'Error: ' + unicode(e), 'info_ru': u'Ошибка: ' + unicode(e)}

    def __checkShopCode(self, shopCode, nodes):
        """Проверка наличия магазина с переданным кодом"""
        if not shopCode or str(shopCode).strip() == '':
            return False, shopCode
        if nodes and not self.shopCodeIsShopId:
            if not nodes.id or not str(shopCode) in nodes.id:
                self.log.warning(
                    u'По коду магазина %s найдено несколько идентификаторов либо ни одного. Невозможно однозначно определить целевой магазин.' % shopCode)
                return False, shopCode
            shopCode = nodes.id
        elif not self.shopCodeIsShopId:
            self.log.warning(u'Невозможно найти по коду магазина его идентификатор, т.к. ноды не считаны')
            return False, shopCode
        return True, shopCode

    def __loggingRequest(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            """Логирование входящего запроса"""
            self.log.info('Получен запрос от %s' % request.remote_addr)
            backuplog = logger.getUploadLogger('RestBackup')
            backuplog.debug('Получен запрос от %s' % request.remote_addr)
            backuplog.debug('remote_addr = %s ' % request.remote_addr)
            backuplog.debug('url = %s ' % request.url)
            backuplog.debug('method = %s ' % request.method)
            backuplog.debug('headers = %s ' % (request.headers.items()))
            backuplog.debug('content_length = %s ' % request.content_length)
            backuplog.debug('chunked = %s ' % request.chunked)
            if request.content_length > 0 or request.chunked:
                backuplog.debug('body = %s ' % request.body.read())
            self.log.debug('Получен запрос от %s' % request.remote_addr)
            try:
                actual_response = fn(*args, **kwargs)
            except Exception as e:
                self.log.error(traceback.format_exc())
                raise e
            self.log.info('Обработка запроса завершена')
            backuplog.debug(actual_response)
            backuplog.debug('Обработка запроса завершена успешно')
            self.__backupRequest(request.query.get('shopcode', 'All'))
            return actual_response

        return wrapper

    def __backupRequest(self, shopCode):
        """Архивирование лога с полученным запросом во избежание разрастания логов"""
        logDir = logger.getLogDirectory()
        logFiles = filesmodify.getFilesByMask(logDir, ['%s.*' % self.restLogName])
        for f in logFiles:
            filesmodify.archive(self.log, os.path.dirname(f), os.path.join(logDir, 'REST_BACKUP', shopCode), maxCountFiles=self.maxCountBackupRequests,
                                actualCountDays=self.actualCountDaysForBackupRequests, removeSRC=True, fileName=os.path.basename(f), loginfo=False)


class EmptyNode(object):

    def __init__(self, isCashRun, shopCode, directLoad=False, statusServer=None, dictionaryId=None):
        self.__readConfig()
        self.id = shopCode
        self.lastTimeReadNodes = 0.0
        self.isCashRun = isCashRun
        self.dictionaryId = dictionaryId
        self.currentConverter = self.RestConverter()
        self.extendedProperties = {}
        self.cashCode = ""
        self.nesServer = self.defaultNesServer
        self.statusServer = statusServer
        self.queuepath = config.getString('FILEQUEUE', 'fileQueueRoot', '/linuxcash/cash/data/queues' if self.isCashRun else '/opt/queues')
        self.inserter = DBInserterConfig() if directLoad else QueueInserterConfig(
            queue=parameters.QueueParameters(self.queuepath, self.pack_size, 'artix.loaddict.cash' if self.isCashRun else None))
        self.lastTimeReadNodes = time.time() - self.timeoutNodeReread
        self.log = logger.getUploadLogger("Node", self.id)
        self.nodeGetter = __import__('upload.node.%s.nodeGetter' % (config.getString('UPLOAD', 'nodeGetter', 'cash' if self.isCashRun else 'cs30')),
                                     fromlist='q')

    def getNode(self):
        # в ноде уже есть статус сервер, но есть вариант когда статус сервер передается в запросе, и ответы необходимо слать на него (Губернский лекарь,
        # загрузска справочников через REST на прямую в кассу)
        self.currentConverter.statusServer = self.defaultStatusServer if not self.statusServer else self.statusServer
        nodes = self.__readNodes()
        if nodes:
            for n in nodes:
                if str(n.id) == str(self.id):
                    self.inserter = n.inserter
                    for converter in n.converters:
                        if converter.name == 'aif':
                            self.currentConverter.hotKeyByBarcode = converter.hotKeyByBarcode
                    break
        self.dictionaryId = 'rest' + str(time.time()) if not self.dictionaryId else self.dictionaryId
        try:
            self.currentConverter.dict_state_id = rest_state.sendDictState(self, 'Start convert dict', u'Начало конвертирования справочника.',
                                                                           rest_state.STATUS_DURING, self.dictionaryId)
            if not self.currentConverter.dict_state_id:
                self.currentConverter.dict_state_id = self.dictionaryId
        except Exception as e:
            self.log.warning(u'Ошибка отправки статуса на КС через REST. %s' % e)
            self.log.warning(traceback.format_exc())
        return self

    def __readNodes(self):
        """чтение нод на кассе/кс"""
        nodes = None
        if time.time() - self.lastTimeReadNodes > self.timeoutNodeReread:
            self.lastTimeReadNodes = time.time()
            try:
                self.log.info(u'Чтение нод')
                self.nodeGetter.defaultNesServer = self.defaultNesServer
                self.nodeGetter.defaultStatusServer = self.defaultStatusServer
                nodes = self.nodeGetter.getNodes()
                self.log.info(u'ноды = %s' % ' '.join(str(e) for e in nodes))
            except Exception as e:
                self.log.warning(u'Ошибка считывания нод. %s' % str(e))
                self.log.warning(traceback.format_exc())
        return nodes

    def __readConfig(self):
        """Чтение настроек"""
        self.shopCodeIsShopId = config.getBool('REST_UPLOAD', 'shopCodeIsShopId', True)
        self.timeoutNodeReread = config.getInt('UPLOAD', 'timeoutNodeReread', 300)
        self.defaultStatusServer = config.getString('CS', 'defaultStatusServer', 'localhost:8080')
        self.defaultNesServer = config.getString('CS', 'defaultNesServer', 'localhost:28081')
        self.pack_size = config.getInt('FILEQUEUE', 'queueChunkSize', 25000)

    class RestConverter(object):
        """Конвертер. Используется только в Node, поэтому вложенный"""
        name = 'rest'
        dict_state_id = None

        def __init__(self, hotKeyByBarcode=False):
            self.hotKeyByBarcode = hotKeyByBarcode
